# -*- coding: utf-8 -*-
# Copyright © 2007 Lateef Alabi-Oki
#
# This file is part of striim.
#
# striim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# striim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with striim; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

"""
This module documents a class that defines the behavior of the treeview
for the station manager window.

@author: Lateef Alabi-Oki
@organization: The Scribes Project
@copyright: Copyright © 2007 Lateef Alabi-Oki
@license: GNU GPLv2 or Later
@contact: mystilleef@gmail.com
"""

class TreeView(object):
	"""
	This class defines the behavior of the treeview for the station
	manager window.
	"""

	def __init__(self, manager, player):
		"""
		Initialize object.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param manager: Reference to the Manager instance.
		@type manager: A Manager object.

		@param player: Reference to the Player instance.
		@type player: A Player object.
		"""
		self.__init_attributes(manager, player)
		self.__set_properties()
		self.__signal_id_1 = manager.connect("destroy", self.__destroy_cb)
		self.__signal_id_2 = self.__treeview.connect("notify::sensitive", self.__sensitive_cb)
		self.__signal_id_3 = self.__treeview.connect("row-activated", self.__row_activated_cb)
		self.__signal_id_4 = self.__treeview.connect("cursor-changed", self.__cursor_changed_cb)
		self.__signal_id_5 = self.__treeview.connect("key-press-event", self.__key_press_event_cb)
		self.__signal_id_6 = manager.connect("delete-selections", self.__delete_selections_cb)
		self.__signal_id_7 = manager.connect("showed-window", self.__showed_window_cb)
		self.__signal_id_8 = manager.connect("station-edited", self.__station_edited_cb)
		self.__signal_id_9 = manager.connect("genre-changed", self.__genre_changed_cb)
		# Monitor database for changes.
		from gnomevfs import monitor_add, MONITOR_FILE
		self.__monitor_id = monitor_add(self.__database_uri, MONITOR_FILE, self.__database_changed_cb)
		self.__monitor_id_1 = monitor_add(self.__database_playlist_uri, MONITOR_FILE, self.__playlist_database_changed_cb)
		self.__populate_model()

	def __init_attributes(self, manager, player):
		"""
		Initialize data attributes.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param manager: Reference to the Manager instance.
		@type manager: A Manager object.

		@param player: Reference to the Player instance.
		@type player: A Player object.
		"""
		self.__edited_station_name = None
		self.__manager = manager
		self.__player = player
		self.__treeview = manager.glade.get_widget("TreeView")
		self.__model = self.__create_model()
		self.__column = self.__create_column()
		self.__genre = None
		# Path to the stations database.
		from Striim.Global import metadata_folder
		self.__database_path = metadata_folder + "/stations.gdb"
		self.__playlist_database_path = metadata_folder + "/playlist.gdb"
		from gnomevfs import get_uri_from_local_path
		self.__database_uri = get_uri_from_local_path(self.__database_path)
		self.__database_playlist_uri = get_uri_from_local_path(self.__playlist_database_path)
		self.__signal_id_1 = self.__signal_id_2 = None
		self.__signal_id_3 = self.__signal_id_4 = None
		self.__monitor_id = self.__monitor_id_1 = None
		return

	def __populate_model(self, GENRE=None):
		"""
		Populate the treeview's model.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param GENRE: A radio station's Genre.
		@type GENRE: A String object.
		"""
		boolean = False
		from operator import not_, ne
		self.__treeview.set_property("sensitive", boolean)
		from Metadata import get_stations_from_database
		self.__model.clear()
		stations = get_stations_from_database()
		if not_(stations): return
		for station_name, value in stations.iteritems():
			user_defined_station_name, genre, uris = value
			if GENRE:
				if ne(genre, GENRE): continue
			if not_(user_defined_station_name):
				user_defined_station_name = station_name
			boolean = True
			self.__model.append([user_defined_station_name, station_name, genre])
		self.__treeview.set_property("sensitive", boolean)
		if not_(boolean): return
		self.__select_row()
		return

	def __select_row(self):
		"""
		Select the first row in the treeview.

		@param self: Reference to the DescriptionTreeView instance.
		@type self: A DescriptionTreeView object.
		"""
		try:
			selection = self.__treeview.get_selection()
			selection.select_path(0)
			iterator = self.__model.get_iter_first()
			path = self.__model.get_path(iterator)
			self.__treeview.scroll_to_cell(path, self.__treeview.get_column(0), True, 0.5, 0.0)
			self.__treeview.set_cursor(path, self.__treeview.get_column(0))
			self.__treeview.columns_autosize()
			self.__treeview.grab_focus()
		except TypeError:
			pass
		return

	def __set_properties(self):
		"""
		Set default properties.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.
		"""
		from gtk import SELECTION_MULTIPLE
		selection = self.__treeview.get_selection()
		selection.set_mode(SELECTION_MULTIPLE)
		self.__treeview.set_property("sensitive", False)
		self.__treeview.append_column(self.__column)
		self.__treeview.set_property("model", self.__model)
		self.__treeview.notify("sensitive")
		self.__column.clicked()
		return

	def __create_model(self):
		"""
		Create a model for the treeview.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@return: A model for the treeview.
		@rtype: A gtk.ListStore object.
		"""
		from gtk import ListStore
		# ListStore: (station_name, user_defined_station_name, genre)
		model = ListStore(str, str, str)
		return model

	def __create_column(self):
		"""
		Create a column for the treeview.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@return: A column for the treeview.
		@rtype: A gtk.TreeView object.
		"""
		from gtk import TreeViewColumn, TREE_VIEW_COLUMN_GROW_ONLY
		from gtk import SORT_ASCENDING, CellRendererText
		from i18n import msg0001
		column = TreeViewColumn(msg0001, CellRendererText(), text=0)
		column.set_property("expand", False)
		column.set_property("sizing", TREE_VIEW_COLUMN_GROW_ONLY)
		column.set_property("clickable", True)
		column.set_sort_column_id(0)
		column.set_property("sort-indicator", True)
		column.set_property("sort-order", SORT_ASCENDING)
		return column

	def __delete_selections(self):
		"""
		Delete selected radio stations.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.
		"""
		self.__treeview.set_property("sensitive", False)
		selection = self.__treeview.get_selection()
		model, paths = selection.get_selected_rows()
		from operator import not_
		if not_(paths): return
		from Metadata import get_stations_from_database
		stations = get_stations_from_database()
		if not_(stations): return
		for path in paths:
			iterator = model.get_iter(path)
			station_name = model.get_value(iterator, 1)
			del stations[station_name]
		from Metadata import update_stations_database
		update_stations_database(stations)
		return

	def __destroy_cb(self, *args):
		"""
		Handles callback when the "destroy" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.
		"""
		from Striim.utils import delete_attributes, disconnect_signal
		disconnect_signal(self.__signal_id_1, self.__manager)
		disconnect_signal(self.__signal_id_2, self.__treeview)
		disconnect_signal(self.__signal_id_3, self.__treeview)
		disconnect_signal(self.__signal_id_4, self.__treeview)
		disconnect_signal(self.__signal_id_5, self.__treeview)
		disconnect_signal(self.__signal_id_6, self.__manager)
		disconnect_signal(self.__signal_id_7, self.__manager)
		disconnect_signal(self.__signal_id_8, self.__manager)
		disconnect_signal(self.__signal_id_9, self.__manager)
		from gnomevfs import monitor_cancel
		if self.__monitor_id: monitor_cancel(self.__monitor_id)
		self.__treeview.destroy()
		delete_attributes(self)
		del self
		self = None
		return

	def __row_activated_cb(self, treeview, path, column):
		"""
		Handles callback when the "row-activated" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param treeview: Reference to the TreeView instance.
		@type treeview: A TreeView object.

		@param path: A object representing a row in the treeview.
		@type path: A Tuple object.

		@param column: An object representing a column of a treeview.
		@type column: A gtk.TreeViewColumn object.
		"""
		iterator = self.__model.get_iter(path)
		station_name = self.__model.get_value(iterator, 1)
		from Metadata import get_stations_from_database
		stations = get_stations_from_database()
		from operator import not_
		if not_(stations): return False
		dictionary = {station_name: stations[station_name]}
		self.__player.update_playlist(dictionary)
		return False

	def __cursor_changed_cb(self, treeview):
		"""
		Handles callback when the "cursor-changed" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@return: True to propagate signals to parent widgets.
		@type: A Boolean Object.
		"""
		selection = treeview.get_selection()
		model, paths = selection.get_selected_rows()
		from operator import not_
		if not_(paths): return False
		iterator = model.get_iter(paths[-1])
		station_name = model.get_value(iterator, 1)
		user_defined_station_name = model.get_value(iterator, 0)
		genre = model.get_value(iterator, 2)
		info = station_name, user_defined_station_name, genre
		self.__manager.emit("treeview-changed", info)
		return False

	def __delete_selections_cb(self, *args):
		"""
		Handles callback when the "delete-selections" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.
		"""
		self.__delete_selections()
		return

	def __key_press_event_cb(self, treeview, event):
		"""
		Handles callback when the "key-press-event" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param treeview: Reference to the TreeView instance.
		@type treeview: A TreeView object.

		@param event: Keystroke events.
		@type event: A gtk.Event object.
		"""
		from operator import ne
		from gtk import keysyms
		if ne(event.keyval, keysyms.Delete): return False
		self.__delete_selections()
		self.__treeview.columns_autosize()
		return True

	def __sensitive_cb(self, treeview, data):
		"""
		Handles callback when the "sensitive" property changes.

		@param self: Reference to the DescriptionTreeView instance.
		@type self: A DescriptionTreeView object.

		@param treeview: Reference to the DescriptionTreeView.
		@type treeview: A DescriptionTreeView object.

		@param data: Random data
		@type data: A GObject object.
		"""
		self.__manager.emit("treeview-sensitivity", treeview.get_property("sensitive"))
		return

	def __database_changed_cb(self, monitor_uri, info_uri, event_type):
		"""
		Handles callback when the abbreviation database is modified.

		@param self: Reference to the TemplateLoader instance.
		@type self: An TemplateLoader object.

		@param monitor_uri: The uri that is monitored.
		@type monitor_uri: A String object.

		@param info_uri: The uri that is monitored.
		@type info_uri: A String object.

		@param event_type: The type of modification that occured.
		@type event_type: A gnomevfs.MONITOR_EVENT* object.
		"""
		from gnomevfs import MONITOR_EVENT_DELETED
		from gnomevfs import MONITOR_EVENT_CREATED
		from gnomevfs import MONITOR_EVENT_CHANGED
		from operator import contains
		events = [MONITOR_EVENT_CHANGED, MONITOR_EVENT_DELETED, MONITOR_EVENT_CREATED]
		if contains(events, event_type):
			self.__populate_model(self.__genre)
			self.__select_edited_row()
		return

	def __playlist_database_changed_cb(self, monitor_uri, info_uri, event_type):
		"""
		Handles callback when the abbreviation database is modified.

		@param self: Reference to the TemplateLoader instance.
		@type self: An TemplateLoader object.

		@param monitor_uri: The uri that is monitored.
		@type monitor_uri: A String object.

		@param info_uri: The uri that is monitored.
		@type info_uri: A String object.

		@param event_type: The type of modification that occured.
		@type event_type: A gnomevfs.MONITOR_EVENT* object.
		"""
		from gnomevfs import MONITOR_EVENT_DELETED
		from gnomevfs import MONITOR_EVENT_CREATED
		from gnomevfs import MONITOR_EVENT_CHANGED
		from operator import contains
		events = [MONITOR_EVENT_CHANGED, MONITOR_EVENT_DELETED, MONITOR_EVENT_CREATED]
		if contains(events, event_type):
			from Metadata import get_playlist_from_database
			station = get_playlist_from_database()
			from operator import not_
			if not_(station): return
			self.__edited_station_name = station.keys()[0]
			self.__select_edited_row()
		return

	def __showed_window_cb(self, *args):
		"""
		Handles callback when the "showed-window" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.
		"""
		self.__treeview.grab_focus()
		self.__treeview.columns_autosize()
		return

	def __station_edited_cb(self, manager, station_name):
		"""
		Handles callback when the "station-edited" signal is emitted.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param manager: Reference to the Manager instance.
		@type manager: A Manager object.

		@param station_name: Name of station that has been edited.
		@type station_name: A String object.
		"""
		self.__edited_station_name = station_name
		return

	def __genre_changed_cb(self, manager, genre):
		from operator import eq
		if eq(genre, "All"): genre = None
		self.__genre = genre
		self.__populate_model(genre)
		return

	def __select_edited_row(self):
		"""
		Select recently edited row.

		@param self: Reference to the TreeView instance.
		@type self: A TreeView object.

		@param station_name: A radio station name.
		@type station_name: A String object.
		"""
		from operator import not_, eq
		if not_(self.__edited_station_name): return
		iterator = self.__model.get_iter_first()
		if not_(iterator): return
		while True:
			name = self.__model.get_value(iterator, 1)
			if eq(name, self.__edited_station_name):
				selection = self.__treeview.get_selection()
				selection.unselect_all()
				selection.select_iter(iterator)
				path = self.__model.get_path(iterator)
				self.__treeview.scroll_to_cell(path, self.__treeview.get_column(0), True, 0.5, 0.0)
				self.__treeview.set_cursor(path, self.__treeview.get_column(0))
				self.__treeview.columns_autosize()
				self.__treeview.grab_focus()
				break
			iterator = self.__model.iter_next(iterator)
			if not_(iterator): break
		self.__edited_station_name = None
		return
