#! /usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright © 2007 Lateef Alabi-Oki
#
# This file is part of striim.
#
# striim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# striim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with striim; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

streamer_dbus_service = "org.sourceforge.striimStreamer"
streamer_dbus_path = "/org/sourceforge/striimStreamer"

class Streamer(object):
	"""
	This class creates an object that plays radio station streams.
	The object is run in a separate process. Communication with this
	process is done via D-Bus.
	"""

	def __init__(self):
		"""
		Initialize object.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		self.__start_up_check()
		from DBusService import DBusService
		dbus_service = DBusService(self)
		from sys import setcheckinterval
		setcheckinterval(10)
		self.__init_attributes(dbus_service)
		self.__set_properties()
		self.__bus.add_signal_watch()
		self.__bus.connect("message::tag", self.__message_tag_cb)
		self.__bus.connect("message::error", self.__message_error_cb)
		self.__bus.connect("message::info", self.__message_info_cb)
		self.__bus.connect("message::warning", self.__message_warning_cb)
		self.__bus.connect("message::buffering", self.__message_buffering_cb)
		self.__bus.connect("message::state-changed", self.__message_state_changed_cb)
		from gnomevfs import monitor_add, MONITOR_FILE
		self.__volume_id = monitor_add(self.__volume_database_uri,
									MONITOR_FILE,
									self.__database_changed_cb)
		self.__reset()
		dbus_service.ready()
		from gobject import idle_add
		idle_add(self.__precompile_methods)

########################################################################
#
#				 		Object Attributes
#
########################################################################

	def __init_attributes(self, dbus_service):
		"""
		Initialize data attributes.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from gst import element_factory_make
		# Path to the abbreviation database.
		from Striim.Global import metadata_folder
		self.__database_path = metadata_folder + "/playlist.gdb"
		self.__volume_database_path = metadata_folder + "/volume.gdb"
		from gnomevfs import get_uri_from_local_path
		self.__database_uri = get_uri_from_local_path(self.__database_path)
		self.__volume_database_uri = get_uri_from_local_path(self.__volume_database_path)
		self.__element = element_factory_make("playbin", "MainPipeline")
		self.__bus = self.__element.get_bus()
		self.__title = None
		self.__uris = set([])
		self.__uri = None
		self.__station = None
		self.__is_playing = False
		self.__metadata_dictionary = {}
		self.__play_time = None
		self.__dbus_service = dbus_service
		return

	def __set_properties(self):
		"""
		Set default properties.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		queue_size = self.__element.get_property("queue-size") * 1000000000
		volume = self.__get_volume_from_database()
		self.__element.set_property("volume", volume)
		self.__element.set_property("queue-size", long(queue_size))
		return

########################################################################
#
#							Public Methods
#
########################################################################

	def toggle_play(self):
		"""
		Start or stop streaming radio station.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from gobject import idle_add
		if self.__is_playing:
			idle_add(self.__pause)
		else:
			idle_add(self.__reset)
		return

	def play_new_station(self):
		"""
		Play a new radio station.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from gobject import idle_add, source_remove
		try:
			source_remove(self.__play_new_station_id)
		except:
			pass
		self.__play_new_station_id = idle_add(self.__reset)
		return

########################################################################
#
#						Control Methods
#
########################################################################

	def __play(self):
		"""
		Start streaming.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from operator import not_
		if not_(self.__uri): return
		self.__dbus_service.busy()
		from gst import STATE_PAUSED
		self.__element.set_state(STATE_PAUSED)
		from time import time
		self.__play_time = time()
		return False

	def __pause(self):
		"""
		Stop streaming radio station.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		self.__stop()
		self.__dbus_service.stopped()
		return False

	def __stop(self):
		"""
		Stop streaming.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		self.__title = None
		self.__metadata_dictionary.clear()
		from gst import STATE_NULL, STATE_READY
		self.__element.set_state(STATE_READY)
		self.__element.set_state(STATE_NULL)
		self.__is_playing = False
		return False

	def __replay(self):
		"""
		Restream radio station.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		self.__stop()
		self.__set_uri()
		from gobject import idle_add
		idle_add(self.__play)
		return False

	def __reset(self):
		"""
		Reset radio station and stream.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		try:
			from Exceptions import PlaylistError
			self.__get_playlist_from_database()
			self.__replay()
		except PlaylistError:
			self.__dbus_service.playlist_error()
		return False

########################################################################
#
#							Helper Methods
#
########################################################################

	def __update_metadata_dictionary(self, message):
		"""
		Update metadata information about a striim.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param message: A message about a striim.
		@type message: A Structure object.
		"""
		taglist = message.parse_tag()
		from operator import truth, not_, contains, eq
		if truth(taglist.is_empty()): return False
		if not_(contains(taglist.keys(), "title")): return False
		if eq(taglist["title"], self.__title): return False
		self.__title = taglist["title"]
		from xml.sax.saxutils import escape
		self.__metadata_dictionary["station"] = escape(self.__station)
		self.__metadata_dictionary["time"] = self.__calculate_time()
		self.__metadata_dictionary["title"] = escape(self.__title.strip())
		self.__dbus_service.stream_update(self.__metadata_dictionary)
		return False

	def __set_uri(self):
		"""
		Set stream source.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		from operator import not_
		if not_(self.__uris): return
		self.__uri = self.__uris.pop()
		self.__element.set_property("uri", self.__uri)
		return

	def __calculate_time(self):
		"""
		Calculate how long the a radio station has been streaming.

		There's got to be a better way to do this, via GStreamer.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param time: How long a station has been playing in hours and minutes.
		@type time: A String object.
		"""
		from time import time
		self.__current_time = time()
		duration = int((self.__current_time - self.__play_time) / 60)
		if duration == 0:
			time = "1min"
		elif duration >= 60:
			hour = str(int(duration / 60))
			mins = str(int(duration % 60))
			if mins == "0":
				time = "%shr" % (hour)
			else:
				time = "%shr %smin" % (hour, mins)
		else:
			time = "%smin" % duration
		return time

	def __get_volume_from_database(self):
		"""
		Get volume from database.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from Metadata import get_volume_from_database
		volume = get_volume_from_database()
		return volume

	def __update_volume(self):
		"""
		Update volume.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		volume = self.__get_volume_from_database()
		self.__element.set_property("volume", volume)
		return False

	def __get_playlist_from_database(self):
		"""
		Get radio station to stream from database.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		from Metadata import get_playlist_from_database
		dictionary = get_playlist_from_database()
		from operator import not_
		from Exceptions import PlaylistError
		if not_(dictionary): raise PlaylistError
		station_name = dictionary.keys()[0]
		user_defined_station_name, genre, self.__uris = dictionary.values()[0]
		if user_defined_station_name:
			self.__station = user_defined_station_name
		else:
			self.__station = station_name
		return

	def __start_up_check(self):
		"""
		Check if another Streamer process is running.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		from Striim.Global import dbus_iface
		services = dbus_iface.ListNames()
		from operator import contains, not_
		if not_(contains(services, streamer_dbus_service)): return
		print "Ooops! Found another streamer, killing this one."
		from os import _exit
		_exit(0)
		return

	def __quit(self):
		"""
		Quit this process.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		self.__stop()
		from os import _exit
		_exit(0)
		return

	def __precompile_methods(self):
		"""
		Use psyco to precompile some methods to improve performance.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.
		"""
		try:
			from psyco import bind
			bind(self.__play)
			bind(self.__pause)
			bind(self.__update_metadata_dictionary)
			bind(self.__get_playlist_from_database)
			bind(self.__set_uri)
			bind(self.__calculate_time)
		except ImportError:
			pass
		return False

########################################################################
#
#					Gstreamer Bus Callback Handlers
#
########################################################################

	def __message_tag_cb(self, bus, message):
		"""
		Handles callback when metadata information about a track is
		found.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param bus: A message system for a GStreamer element.
		@type bus: A GStreamer object.

		@param message: Some data.
		@type message: A data structure object.
		"""
		from gobject import idle_add
		idle_add(self.__update_metadata_dictionary, message)
		return True

	def __message_error_cb(self, bus, message):
		"""
		Handles callback when a stream error occurs.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param bus: A message system for a GStreamer element.
		@type bus: A GStreamer object.

		@param message: Some data.
		@type message: A data structure object.
		"""
		error, debug = message.parse_error()
		if self.__uris:
			self.__replay()
		else:
			self.__is_playing = False
			from xml.sax.saxutils import escape
			self.__dbus_service.stream_error(escape(str(error)))
		return True

	def __message_buffering_cb(self, bus, message):
		"""
		Handles callback when a stream is buffering.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param bus: A message system for a GStreamer element.
		@type bus: A GStreamer object.

		@param message: Some data.
		@type message: A data structure object.
		"""
		from gst import STATE_PAUSED, STATE_PLAYING
		percentage = message.structure["buffer-percent"]
		from operator import eq
		if eq(percentage, 100):
			self.__element.set_state(STATE_PLAYING)
			self.__dbus_service.stream_message("done")
		else:
			self.__element.set_state(STATE_PAUSED)
			from i18n import msg0002
			string = msg0002 % percentage
			from xml.sax.saxutils import escape
			self.__dbus_service.stream_message(escape(string))
		return True

	def __message_warning_cb(self, bus, message):
		"""
		Handles callback when a warning occurs.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param bus: A message system for a GStreamer element.
		@type bus: A GStreamer object.

		@param message: Some data.
		@type message: A data structure object.
		"""
		warning = message.parse_warning()
		print "Gstreamer Warning: " + warning
		return True

	def __message_info_cb(self, bus, message):
		"""
		Handles callback when a info occurs.

		@param self: Reference to the Manager instance.
		@type self: A Manager object.

		@param bus: A message system for a GStreamer element.
		@type bus: A GStreamer object.

		@param message: Some data.
		@type message: A data structure object.
		"""
		info = message.parse_info()
		print "Gstreamer Info: " + info
		return True

	def __message_state_changed_cb(self, bus, data):
		state = data.parse_state_changed()
		from gst import STATE_PLAYING
		from operator import eq
		if eq(state[1], STATE_PLAYING):
			self.__is_playing = True
		else:
			self.__is_playing = False
		return True

########################################################################
#
#					Database Monitor Callback Handler
#
########################################################################

	def __database_changed_cb(self, monitor_uri, info_uri, event_type):
		"""
		Handles callback when the playlist database is modified.

		@param self: Reference to the Manager instance.
		@type self: An Manager object.

		@param monitor_uri: The uri that is monitored.
		@type monitor_uri: A String object.

		@param info_uri: The uri that is monitored.
		@type info_uri: A String object.

		@param event_type: The type of modification that occured.
		@type event_type: A gnomevfs.MONITOR_EVENT* object.
		"""
		from gnomevfs import MONITOR_EVENT_DELETED
		from gnomevfs import MONITOR_EVENT_CREATED
		from gnomevfs import MONITOR_EVENT_CHANGED
		if event_type in [MONITOR_EVENT_DELETED, MONITOR_EVENT_CREATED, MONITOR_EVENT_CHANGED]:
			from gobject import source_remove, idle_add
			from operator import eq
			if eq(monitor_uri, self.__database_uri):
				try:
					source_remove(self.__reset_id)
				except:
					pass
				self.__reset_id = idle_add(self.__reset)
			else:
				try:
					source_remove(self.__update_volume_id)
				except:
					pass
				self.__update_volume_id = idle_add(self.__update_volume)
		return

########################################################################
#
#					GUI process	D-Bus Callback Handler
#
########################################################################

	def __name_change_cb(self, *args):
		"""
		Handles callback when the striim GUI process dies.

		@param self: Reference to the Streamer instance.
		@type self: A Streamer object.
		"""
		self.__quit()
		return

########################################################################
#
#							Debug Methods
#
########################################################################

	def __print_info(self):
		"""
		Get necessariy information about system Scribes is running on.

		@param self: Reference to the CommandLineProcessor instance.
		@type self: A CommandLineProcessor object.
		"""
		print "========================================================"
		import os
		print "System Info: ", os.uname()
		import sys
		print "Python Version: ", sys.version
		print "System Byteorder: ", sys.byteorder
		print "Python Modules: ", sys.builtin_module_names
		print "========================================================"
		from Striim.Global import version
		print "striim Version: ", version
		import dbus
		print "Dbus Version: ", dbus.version
		import gtk
		print "GTK+ Version: ", gtk.gtk_version
		import gtk
		print "PyGTK Version: ",	gtk.pygtk_version
		import gnome
		print "GNOME Python Version: ", gnome.gnome_python_version
		import gst
		print "Gstreamer Version: ", gst.version()
		try:
			import psyco
			print "Psyco Version: ", psyco.version_info
		except ImportError:
			print "Psyco Not Installed"
		print "========================================================"
		from Striim.Global import dbus_iface
		services = dbus_iface.ListNames()
		service = "net.sourceforge.striim"
		print "Running Instance: ", services.count(service)
		service = "net.sourceforge.striimStreamer"
		print "Running Streamer Instance: ", services.count(service)
		print "========================================================"
		from Striim.Global import striim_executable_path
		from Striim.Global import python_path, core_plugin_folder
		from Striim.Global import striim_data_folder
		print "Python Path: ", python_path
		print "Plugin Path: ", core_plugin_folder
		print "Data Path: ", striim_data_folder
		print "Executable Path: ", striim_executable_path
		print "========================================================"
		return

if __name__ == "__main__":
	from sys import argv, path
	python_path = argv[1]
	path.insert(0, python_path)
	from gobject import threads_init, MainLoop
	threads_init()
	Streamer()
	MainLoop().run()
